
// This demo shows how to use a VT as kind of progress info component. The
// versatility of VT allows for easy look and feel adjustments, and with custom
// drawing even embedded progress bars should be no problem.
//
// Advantage of this unusual method: full separation of progress steps and
// presentation, number of steps and their order can be changed very easily without
// GUI adjustments.
//
// Please note that this demo is not complete, it allows for only one run,
// cannot be stopped in between and the application may not be very responsive
// during one run.
//
// For your information, in my original implementation the form with this progress
// view implements the following interface. This makes it easy to have different
// code portions perform tasks and let them display the progress to the user without
// having to know how this is done exactly.
//
// type
//   IProgress = interface
//     function GetStage: TProgressStage;
//     function GetStageProgress(Stage: TProgressStage): string;
//     function GetStageInfo(Stage: TProgressStage): string;
//     procedure SetStage(const Value: TProgressStage);
//     procedure SetStageProgress(Stage: TProgressStage; const Value: string);
//     procedure SetStageInfo(Stage: TProgressStage; const Value: string);
//
//     procedure UpdateGUI(Force: Boolean = false; Node: PVirtualNode = nil);
//
//     property Stage: TProgressStage read GetStage write SetStage;
//     property StageInfo[Stage: TProgressStage]: string read GetStageInfo write SetStageInfo;
//     property StageProgress[Stage: TProgressStage]: string read GetStageProgres write SetStageProgress;
//   end;
//
//
// Werner Lehmann <wl@bwl.uni-kiel.de>, 18-Nov-2001.

unit Unit1;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, VirtualTrees, ExtCtrls, StdCtrls, ImgList;

type
  TProgressStage = (psInit, psPreProc, psDataAlloc, psAnalyze);

  TProgressInfo = record
    Info, Progress: string;
    Node: PVirtualNode;
  end;

const
  ProgressStageName: array[TProgressStage] of string =
      ('Initialization', 'Preprocessing', 'Data Allocation', 'Analyzation');

type
  TForm1 = class(TForm)
    imProgress: TImageList;
    bStart: TButton;
    Panel1: TPanel;
    VT: TVirtualStringTree;
    procedure FormCreate(Sender: TObject);
    procedure VTGetText(Sender: TBaseVirtualTree; Node: PVirtualNode;
      Column: Integer; TextType: TVSTTextType; var Text: WideString);
    procedure VTGetImageIndex(Sender: TBaseVirtualTree; Node: PVirtualNode;
      Kind: TVTImageKind; Column: Integer; var Ghosted: Boolean;
      var Index: Integer);
    procedure VTPaintText(Sender: TBaseVirtualTree; const Canvas: TCanvas;
      Node: PVirtualNode; Column: Integer; TextType: TVSTTextType);
    procedure bStartClick(Sender: TObject);
  private
    FProgressInfo: array[TProgressStage] of TProgressInfo;
    FStage: TProgressStage;
    FLastUpdate, FLastMessages: DWORD;
    function GetStage: TProgressStage;
    function GetStageInfo(Stage: TProgressStage): string;
    function GetStageProgress(Stage: TProgressStage): string;
    procedure SetStage(const Value: TProgressStage);
    procedure SetStageInfo(Stage: TProgressStage; const Value: string);
    procedure SetStageProgress(Stage: TProgressStage; const Value: string);
  public
    procedure UpdateGUI(Force: Boolean = false; Node: PVirtualNode = nil);

    property Stage: TProgressStage read GetStage write SetStage;
    property StageInfo[Stage: TProgressStage]: string read GetStageInfo write SetStageInfo;
    property StageProgress[Stage: TProgressStage]: string read GetStageProgress write SetStageProgress;
  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}

//------------------------------------------------------------------------------
procedure TForm1.FormCreate(Sender: TObject);
var
  CurStage: TProgressStage;
  Node: PVirtualNode;
begin
  VT.RootNodeCount := Length(ProgressStageName);
  Node := VT.GetFirst;
  for CurStage := Low(TProgressStage) to High(TProgressStage) do
  begin
    FProgressInfo[CurStage].Node := Node;
    FProgressInfo[CurStage].Progress := '';
    FProgressInfo[CurStage].Info := '';
    // VT.IsVisible[Node] := Stage = Low(TProgressStage);
    Node := Node.NextSibling;
  end;
  VT.Invalidate;
end;

//------------------------------------------------------------------------------
procedure TForm1.VTGetText(Sender: TBaseVirtualTree; Node: PVirtualNode;
  Column: Integer; TextType: TVSTTextType; var Text: WideString);
var
  NodeStage: TProgressStage;
begin
  NodeStage := TProgressStage(Node.Index);
  case Column of
    0:
      Text := ProgressStageName[NodeStage];
    1:
      if not bStart.Enabled and (FStage >= NodeStage) and (FProgressInfo[NodeStage].Progress = '') then
        Text := ''
      else
        Text := FProgressInfo[NodeStage].Progress;
    2:
      Text := FProgressInfo[NodeStage].Info;
  end;
end;

//------------------------------------------------------------------------------
procedure TForm1.VTGetImageIndex(Sender: TBaseVirtualTree;
  Node: PVirtualNode; Kind: TVTImageKind; Column: Integer;
  var Ghosted: Boolean; var Index: Integer);
begin
  if (Column = 0) and (Kind = ikSelected) then
    Index := 0;
end;

//------------------------------------------------------------------------------
procedure TForm1.VTPaintText(Sender: TBaseVirtualTree;
  const Canvas: TCanvas; Node: PVirtualNode; Column: Integer;
  TextType: TVSTTextType);
begin
  if Sender.Selected[Node] then
  begin
    Canvas.Font.Color := clBlack;
    if Column = 0 then
      Canvas.Font.Style := [fsBold];
  end;
end;

//------------------------------------------------------------------------------
procedure TForm1.bStartClick(Sender: TObject);
var
  i: Integer;
begin
  bStart.Enabled := false;

  Stage := Low(TProgressStage);
  repeat
    i := 0;
    while i <= 100 do
    begin
      StageProgress[Stage] := IntToStr(i) + '%';
      Sleep(50); // Heavy processing here.
      Inc(i);
    end;
    StageInfo[Stage] := 'completed, fubar, detail info whatever';
    if Stage = High(TProgressStage) then // Cannot increment FStage beyond its domain...
      break;
    Stage := Succ(Stage);
  until false;
  VT.ClearSelection;
end;

//------------------------------------------------------------------------------
// This is an interface method in my original implementation hence no direct
// field getter.
function TForm1.GetStage: TProgressStage;
begin
  result := FStage;
end;

//------------------------------------------------------------------------------
function TForm1.GetStageInfo(Stage: TProgressStage): string;
begin
  result := FProgressInfo[Stage].Info;
end;

//------------------------------------------------------------------------------
function TForm1.GetStageProgress(Stage: TProgressStage): string;
begin
  result := FProgressInfo[Stage].Progress;
end;

//------------------------------------------------------------------------------
procedure TForm1.SetStage(const Value: TProgressStage);
begin
  FStage := Value;
  // VT.IsVisible[FProgressInfo[FStage].Node] := true;
  VT.Selected[FProgressInfo[FStage].Node] := true;
  UpdateGUI(true);
end;

//------------------------------------------------------------------------------
procedure TForm1.SetStageInfo(Stage: TProgressStage; const Value: string);
begin
  FProgressInfo[Stage].Info := Value;
  VT.ReinitNode(FProgressInfo[Stage].Node, false);
  UpdateGUI(false, FProgressInfo[Stage].Node);
end;

//------------------------------------------------------------------------------
procedure TForm1.SetStageProgress(Stage: TProgressStage; const Value: string);
begin
  FProgressInfo[Stage].Progress := Value;
  UpdateGUI(false, FProgressInfo[Stage].Node);
end;

//------------------------------------------------------------------------------
procedure TForm1.UpdateGUI(Force: Boolean; Node: PVirtualNode);
begin
  if Force or (GetTickCount - FLastUpdate >= 300) then
  begin
    FLastUpdate := GetTickCount;
    if Node <> nil then
      VT.InvalidateNode(Node)
    else
      VT.Invalidate;

    if Force or (FLastUpdate - FLastMessages >= 1000) then
    begin
      FLastMessages := FLastUpdate;
      Application.ProcessMessages;
    end else
    begin
      VT.Update;
    end;
  end;
end;

end.

