Unit ZipTree;
{kanca@attglobal.net
 Last Modifications Date : 01.Mar.2000
 Version 1.01
 TZipTree - Collection with hierarchical structure.
 TCollection and TCollectionItem united in one Class
 Simpler to use than a collection. Can contain a collection in a
 collection. Is compatible with TTreeNodes class.
 Sorting and Recursive Sorting; Loading/Saving From/To File, Stream, Strings}

Interface

Uses
  Windows,
  Messages,
  SysUtils,
  Classes;

Type
  TCheckType = (
    CtNone,
    CtCheckBox,
    CtCheckBoxGrayed,
    CtRadioButton
    );

  EZipTreeError = Class(Exception);

  TZipTree = Class;

  TCompareEvent = Function(AItem1, AItem2: TZipTree): Integer;
  TOnChangeEvent = Procedure(AItem: TZipTree) Of Object;

  TZipTree = Class
  Private
    FBookmark: Integer;
    FFileName: String;
    FCheckType: TCheckType;
    FChecked: Boolean;
    FContent: TObject; //An object can be stored here
    FContentType: Integer;
    FData: Pointer;
    FDescendantCount: Integer; //Total number of items all the way under the node --grandchildren and all..
    FFullPath: String;
    FGrayed: Boolean;
    FItems: TList;
    FLevelNo: Integer;
    FOnChange: TOnChangeEvent;
    FParent: TZipTree;
    FPath: String;
    FTag: Integer;
    FUpdateCount: Integer;
    Function GetCount: Integer;
    Function GetIndex: Integer;
    Function HasChild: Boolean;
    Procedure ParseFullPath(AInput: String; AList: TStrings);
    Procedure DescendantCountDecrement(AValue: Integer);
    Procedure DescendantCountIncrement(AValue: Integer);
    Procedure InsertItem(AItem: TZipTree);
    Procedure RemoveItem(AItem: TZipTree);
    Procedure SetChecked(Value: Boolean);
    Procedure SetGrayed(Value: Boolean);
    Procedure SetOnChange(AValue: TOnChangeEvent);
    Procedure SetParent(AParent: TZipTree);
    Procedure UpdateParents(AllItems: Boolean);
  Protected
    Function GetItem(Index: Integer): TZipTree;
    Procedure DefineProperties(Filer: TFiler);
    Procedure ReadData(Reader: TReader); Virtual;
    Procedure SetIndex(Value: Integer); Virtual;
    Procedure SetItem(Index: Integer; Value: TZipTree);
    Procedure Update(AItem: TZipTree); Virtual;
    Procedure WriteData(Writer: TWriter); Virtual;
  Public
    Constructor Create(AParent: TZipTree);
    Destructor Destroy; Override;
    Function Add(AFullPath: String): TZipTree;
    Function FindBookMark(ABookmark: Integer): TZipTree;
    Function GetFirstChild: TZipTree;
    Function GetLastChild: TZipTree;
    Function GetNext: TZipTree;
    Function GetNextSibling: TZipTree;
    Function GetPrev: TZipTree;
    Function GetPrevSibling: TZipTree;
    Function Insert(Index: Integer): TZipTree;
    Function SearchForCaption(AValue: String): TZipTree;
    Procedure AlphaSort(AIsRecursive: Boolean);
    Procedure Assign(Source: TObject);
    Procedure BeginUpdate; Virtual;
    Procedure Clear;
    Procedure CustomSort(ASortProc: TCompareEvent; AIsRecursive: Boolean);
    Procedure DeleteItem(AItem: TZipTree);
    Procedure EndUpdate; Virtual;
    Procedure Exchange(AIndex1, AIndex2: Integer);
    Procedure Sort(ASortProc: TCompareEvent; AIsRecursive: Boolean);
    Property Bookmark: Integer Read FBookmark;
    Property FileName: String Read FFileName Write FFileName;
    Property CheckType: TCheckType Read FCheckType Write FCheckType;
    Property Checked: Boolean Read FChecked Write SetChecked;
    Property Content: TObject Read FContent Write FContent;
    Property ContentType: Integer Read FContentType Write FContentType; //for different types of Collections
    Property Count: Integer Read GetCount;
    Property Data: Pointer Read FData Write FData;
    Property DescendantCount: Integer Read FDescendantCount; //Total number of items under node --grandchildren and all..
    Property FullPath: String Read FFullPath;
    Property Grayed: Boolean Read FGrayed Write SetGrayed;
    Property HasChildren: Boolean Read HasChild;
    Property Index: Integer Read GetIndex Write SetIndex;
    Property Items[Index: Integer]: TZipTree Read GetItem Write SetItem; Default;
    Property Level: Integer Read FLevelNo;
    Property OnChange: TOnChangeEvent Read FOnChange Write SetOnChange;
    Property Parent: TZipTree Read FParent Write SetParent;
    Property Path: String Read FPath;
    Property Tag: Integer Read FTag Write FTag;
  End;

Implementation

Uses
  Consts,
  ComStrs,
  TypInfo,
  Dialogs;

Const
  Version = 1.0;
  HeaderString = 'TZipTree Binary File';
  SzInteger = SizeOf(Integer);
  SzLargeInteger = SizeOf(Int64);
  SzBoolean = SizeOf(Boolean);
  SzSingle = SizeOf(Single);
  SzDouble = SizeOf(Double);
  SzByte = SizeOf(Byte);

Procedure TreeCollectionError(Msg: String);
Begin
  Raise EZipTreeError.Create(Msg);
End;

Procedure QuickSort(SortList: TList; ALeft, ARight: Integer; ACompResult: TCompareEvent);
Var
  TI1: Integer;
  TI2: Integer;
  TP1: Pointer;
  TP2: Pointer;
Begin
  Repeat
    TI1 := ALeft;
    TI2 := ARight;
    TP1 := SortList[(ALeft + ARight) Shr 1];
    Repeat
      While ACompResult(TZipTree(SortList[TI1]), TZipTree(TP1)) < 0 Do Inc(TI1);
      While ACompResult(TZipTree(SortList[TI2]), TZipTree(TP1)) > 0 Do Dec(TI2);
      If TI1 <= TI2 Then Begin
        TP2 := SortList[TI1];
        SortList[TI1] := SortList[TI2];
        SortList[TI2] := TP2;
        Inc(TI1);
        Dec(TI2);
      End;
    Until TI1 > TI2;
    If ALeft < TI2 Then QuickSort(SortList, ALeft, TI2, ACompResult);
    ALeft := TI1;
  Until TI1 >= ARight;
End;

Function DefaultSort(AItem1, AItem2: TZipTree): Integer;
Begin
  Result := LStrcmp(PChar(AItem1.FullPath), PChar(AItem2.FullPath));
End;

Constructor TZipTree.Create(AParent: TZipTree);
Begin
  FParent := Nil;
  FContentType := 0;
  FLevelNo := 0;
  FTag := 0;
  FOnChange := Nil;
  FUpdateCount := 0;
  FGrayed := False;
  FChecked := False;
  FCheckType := CtNone;
  FBookmark := 0;
  FDescendantCount := 0;
  FContent := Nil;
  FData := Nil;
  FFileName := '';
  FItems := TList.Create;
  If AParent <> Nil Then SetParent(AParent); //treecollitem
End;

Destructor TZipTree.Destroy;
Begin
  FUpdateCount := 0;
  If FItems <> Nil Then Clear;
  FItems.Free;
  FData := Nil;
  FContent := Nil;
  SetParent(Nil);
  Inherited Destroy;
End;

Procedure TZipTree.ParseFullPath(AInput: String; AList: TStrings);
  Function Trim(S: String): String;
  Var
    Length1: Integer;
  Begin
    Length1 := Length(S);
    If Length1 > 0 Then Begin
      While (Length1 > 0) And (S[1] <= ' ') Do Begin
        System.Delete(S, 1, 1);
        Length1 := Length(S);
      End;
      Length1 := Length(S);
      While (Length1 > 0) And (S[Length1] <= ' ') Do Begin
        System.Delete(S, Length1, 1);
        Length1 := Length(S);
      End;
    End;
    Result := S;
  End;
Const
  Delimiter = '\';
  DelimiterLen = Length(Delimiter);
Var
  DelimiterPos1: Integer;
  TempStr: String;
Begin
  AList.Clear;
  Repeat
    DelimiterPos1 := Pos(Delimiter, AInput);
    If DelimiterPos1 = 0 Then TempStr := AInput
    Else TempStr := Copy(AInput, 1, DelimiterPos1 - 1);
    If (TempStr <> '') Then AList.Add(Trim(TempStr));
    AInput := Copy(AInput, DelimiterPos1 + DelimiterLen, Length(AInput) - DelimiterPos1);
  Until DelimiterPos1 = 0;
End;

Function TZipTree.SearchForCaption(AValue: String): TZipTree;
  Function TraverseAndSearch(AZipTree: TZipTree): TZipTree;
  Var
    TI1: Integer;
  Begin
    If AZipTree.FullPath = AValue Then Result := AZipTree
    Else Begin
      Result := Nil;
      For TI1 := 0 To AZipTree.Count - 1 Do Begin
        Result := TraverseAndSearch(AZipTree.Items[TI1]);
        If Result <> Nil Then Break;
      End;
    End;
  End;
Var
  ZipTree1: TZipTree;
Begin
  ZipTree1 := Self;
  While Not ((ZipTree1 = Nil) Or (ZipTree1.FParent = Nil)) Do ZipTree1 := ZipTree1.FParent;
  Result := TraverseAndSearch(ZipTree1);
End;

Function TZipTree.Add(AFullPath: String): TZipTree;
  Function FindLastParent(AZipTree: TZipTree; AList: TStrings): TZipTree;
  Var
    tI1: Integer;
  Begin
    Result := AZipTree;
    If AList.Count <= 0 Then Exit;
    If AZipTree.Count <= 0 Then Exit;

    For tI1 := 0 To AZipTree.Count - 1 Do Begin
      If AZipTree[tI1].FileName = AList[0] Then Begin
        AList.Delete(0);
        If AList.Count > 1 Then Result := FindLastParent(AZipTree[tI1], AList)
        Else Exit;
      End;
    End;
  End;

  Function AddEndLeaf(AZipTree: TZipTree; AList: TStrings): TZipTree;
  Begin
    Result := AZipTree;
    If AList.Count <= 0 Then Exit;
    If AZipTree.Count <= 0 Then Exit;
    Result := TZipTree.Create(AZipTree);
    Result.FFileName := AList[0];
    Result.FLevelNo := FLevelNo + 1;
    If Assigned(FOnChange) Then Begin
      Result.OnChange := FOnChange;
      FOnChange(Result);
    End;
    AList.Delete(0);
    Result := AddEndLeaf(Result, AList);
  End;

Var
  PathList1: TStrings;
  ZipTree1: TZipTree;
Begin
  PathList1 := TStringList.Create;
  ParseFullPath(AFullPath, PathList1);
  ZipTree1 := FindLastParent(Self, PathList1);
  ZipTree1 := AddEndLeaf(ZipTree1, PathList1);

  Result := TZipTree.Create(ZipTree1);
  Result.FFileName := PathList1[0];
  Result.FFullPath := AFullPath;
  Result.FLevelNo := FLevelNo + 1;
  If Assigned(FOnChange) Then Begin
    Result.OnChange := FOnChange;
    FOnChange(Result);
  End;
  PathList1.Free;
End;

Procedure TZipTree.DeleteItem(AItem: TZipTree);
Begin
  BeginUpdate;
  RemoveItem(AItem);
  AItem.Clear;
  AItem.Free;
  EndUpdate;
End;

Procedure TZipTree.Assign(Source: TObject);
Var
  TI1: Integer;
Begin
  If Source Is TZipTree Then Begin
    BeginUpdate;
    Try
      FTag := TZipTree(Source).FTag;
      FFileName := TZipTree(Source).FFileName;
      FContentType := TZipTree(Source).FContentType;
      FChecked := TZipTree(Source).FChecked;
      FGrayed := TZipTree(Source).FGrayed;
      Clear;
      For TI1 := 0 To TZipTree(Source).Count - 1 Do Begin
        Add('').Assign(TZipTree(Source).Items[TI1]);
      End;
    Finally
      EndUpdate;
    End;
    Exit;
  End;
End;

Procedure TZipTree.BeginUpdate;
Begin
  Inc(FUpdateCount);
End;

Procedure TZipTree.Update(AItem: TZipTree);
Begin
  If Assigned(FOnChange) Then Begin
    If AItem <> Nil Then FOnChange(AItem)
    Else If FParent <> Nil Then FOnChange(FParent);
  End;
End;

Procedure TZipTree.UpdateParents(AllItems: Boolean);
Begin //update owner collection
  If (FParent <> Nil) And (FParent.FUpdateCount = 0) Then Begin
    If AllItems Then FParent.Update(Nil)
    Else FParent.Update(Self);
  End;
End;

Procedure TZipTree.Clear;
Begin
  If FItems.Count > 0 Then Begin
    BeginUpdate;
    Try
      While FItems.Count > 0 Do TZipTree(FItems.Last).Free;
    Finally
      EndUpdate;
    End;
  End;
End;

Procedure TZipTree.EndUpdate;
Begin
  Dec(FUpdateCount);
  UpdateParents(False);
End;

Function TZipTree.FindBookMark(ABookmark: Integer): TZipTree;
Var
  TI1: Integer;
Begin
  Result := Nil;
  For TI1 := 0 To FItems.Count - 1 Do Begin
    Result := TZipTree(FItems[TI1]);
    If Result.FBookmark = ABookmark Then Exit;
  End;
End;

Function TZipTree.GetCount: Integer;
Begin
  Result := FItems.Count;
End;

Function TZipTree.GetItem(Index: Integer): TZipTree;
Begin
  Result := Nil;
  If (Index < FItems.Count) And (Index >= 0) Then Result := FItems[Index];
End;

Function TZipTree.Insert(Index: Integer): TZipTree;
Begin
  Result := Add('');
  Result.Index := Index;
End;

Procedure TZipTree.InsertItem(AItem: TZipTree);
Begin
  Inc(FBookmark);
  FItems.Add(AItem);
  AItem.FParent := Self;
  DescendantCountIncrement(AItem.FDescendantCount + 1);
  UpdateParents(False);
End;

Procedure TZipTree.RemoveItem(AItem: TZipTree);
Begin
  DescendantCountDecrement(AItem.FDescendantCount + 1);
  FItems.Remove(AItem);
  AItem.FParent := Nil;
  UpdateParents(False);
End;

Procedure TZipTree.DescendantCountIncrement(AValue: Integer);
Begin
  If FParent <> Nil Then FParent.DescendantCountIncrement(AValue);
  FDescendantCount := FDescendantCount + AValue;
End;

Procedure TZipTree.DescendantCountDecrement(AValue: Integer);
Begin
  If FParent <> Nil Then FParent.DescendantCountDecrement(AValue);
  FDescendantCount := FDescendantCount - AValue;
End;

Procedure TZipTree.SetItem(Index: Integer; Value: TZipTree);
Begin
  If (Index < FItems.Count) And (Index >= 0) Then TZipTree(FItems[Index]).Assign(Value)
  Else TreeCollectionError('Index out of bounds');
End;

Procedure TZipTree.SetGrayed(Value: Boolean);
Var
  TI1: Integer;
Begin
  If FParent <> Nil Then Begin
    If FParent.FGrayed Then Begin
      If Value Then FChecked := False;
      FGrayed := Value;
      For TI1 := 0 To Count - 1 Do Items[TI1].Grayed := FGrayed;
    End Else If FGrayed Then Begin
      FGrayed := False;
      For TI1 := 0 To Count - 1 Do Items[TI1].Grayed := FGrayed;
    End;
  End Else Begin
    If Value Then FChecked := False;
    FGrayed := Value;
    For TI1 := 0 To Count - 1 Do Items[TI1].Grayed := FGrayed;
  End;
End;

Procedure TZipTree.SetChecked(Value: Boolean);
Var
  TI1: Integer;
Begin
  If FParent <> Nil Then Begin
    If FParent.FChecked Then Begin
      If Value Then FGrayed := False;
      FChecked := Value;
      For TI1 := 0 To Count - 1 Do Items[TI1].Checked := FChecked;
    End Else If FChecked Then Begin
      FChecked := False;
      For TI1 := 0 To Count - 1 Do Items[TI1].Checked := FChecked;
    End;
  End Else Begin
    If Value Then FGrayed := False;
    FChecked := Value;
    For TI1 := 0 To Count - 1 Do Items[TI1].Checked := FChecked;
  End;
End;

Procedure TZipTree.SetParent(AParent: TZipTree);
Begin
  If FParent <> Nil Then FParent.RemoveItem(Self);
  If AParent <> Nil Then AParent.InsertItem(Self);
End;

Procedure TZipTree.Exchange(AIndex1, AIndex2: Integer);
Begin
  If (AIndex1 >= 0) And (AIndex2 >= 0) And (AIndex1 < Count) And (AIndex2 < Count) Then Begin
    FItems.Exchange(AIndex1, AIndex2);
    UpdateParents(True);
  End;
End;

Function TZipTree.GetIndex: Integer;
Begin
  Result := -1;
  If FParent <> Nil Then Result := FParent.FItems.IndexOf(Self);
End;

Procedure TZipTree.SetIndex(Value: Integer);
Var
  TI1: Integer;
Begin
  TI1 := -1;
  If FParent <> Nil Then TI1 := FParent.FItems.IndexOf(Self);
  If (TI1 >= 0) And (TI1 <> Value) Then Begin
    FParent.FItems.Move(TI1, Value);
    UpdateParents(True);
  End;
End;

Function TZipTree.HasChild: Boolean;
Begin
  Result := FItems.Count > 0;
End;

Function TZipTree.GetFirstChild: TZipTree;
Begin
  Result := Nil;
  If FItems.Count > 0 Then Result := Items[0];
End;

Function TZipTree.GetLastChild: TZipTree;
Begin
  Result := Nil;
  If FItems.Count > 0 Then Result := Items[FItems.Count - 1];
End;

Function TZipTree.GetNext: TZipTree;
Var
  TTC1: TZipTree;
  TTC2: TZipTree; //parent
Begin
  TTC1 := GetFirstChild;
  If TTC1 = Nil Then TTC1 := GetNextSibling;
  TTC2 := FParent; //parent
  While (TTC1 = Nil) And (TTC2 <> Nil) Do Begin
    TTC2 := TTC2.FParent;
    TTC1 := TTC2.GetNextSibling;
  End;
  Result := TTC1;
End;

Function TZipTree.GetPrev: TZipTree;
Var
  TTC1: TZipTree;
Begin
  Result := GetPrevSibling;
  If Result <> Nil Then Begin
    TTC1 := Result;
    Repeat
      Result := TTC1;
      TTC1 := Result.GetLastChild;
    Until TTC1 = Nil;
  End Else Result := FParent;
End;

Function TZipTree.GetNextSibling: TZipTree;
Begin
  Result := FParent.Items[Index + 1];
End;

Function TZipTree.GetPrevSibling: TZipTree;
Var
  TI1: Integer;
Begin
  Result := Nil;
  If FParent <> Nil Then Begin
    TI1 := FParent.FItems.IndexOf(Self);
    If TI1 > 0 Then Result := FParent.Items[TI1 - 1];
  End;
End;

Procedure TZipTree.ReadData(Reader: TReader);
Var
  TTC1: TZipTree;
Begin
  Reader.ReadListBegin;
  BeginUpdate;
  Try
    If Not Reader.EndOfList Then Clear;
    FTag := Reader.ReadInteger;
    FFileName := Reader.ReadString;
    FContentType := Reader.ReadInteger;
    FChecked := Reader.ReadBoolean;
    FGrayed := Reader.ReadBoolean;

    While Not Reader.EndOfList Do Begin
      TTC1 := Add('');
      While Not Reader.EndOfList Do TTC1.ReadData(Reader);
    End;
    Reader.ReadListEnd;
  Finally
    EndUpdate;
  End;
End;

Procedure TZipTree.WriteData(Writer: TWriter);
Var
  TI1: Integer;
Begin
  Writer.WriteListBegin;
  Writer.WriteInteger(FTag);
  Writer.WriteString(FFileName);
  Writer.WriteInteger(FContentType);
  Writer.WriteBoolean(FChecked);
  Writer.WriteBoolean(FGrayed);
  For TI1 := 0 To Count - 1 Do Items[TI1].WriteData(Writer);
  Writer.WriteListEnd;
End;

Procedure TZipTree.DefineProperties(Filer: TFiler);
Begin
  Filer.DefineProperty('TreeCollection', ReadData, WriteData, Count > 1);
End;

Procedure TZipTree.Sort(ASortProc: TCompareEvent; AIsRecursive: Boolean);
Var
  TI1: Integer;
Begin
  BeginUpdate;
  Try
    If (FItems <> Nil) And (Count > 0) Then QuickSort(FItems, 0, Count - 1, ASortProc);
    If AIsRecursive Then For TI1 := 0 To Count - 1 Do Items[TI1].Sort(ASortProc, AIsRecursive);
  Finally
    EndUpdate;
  End;
End;

Procedure TZipTree.CustomSort(ASortProc: TCompareEvent; AIsRecursive: Boolean);
Begin
  If Not Assigned(ASortProc) Then Sort(@DefaultSort, AIsRecursive)
  Else Sort(@ASortProc, AIsRecursive);
End;

Procedure TZipTree.AlphaSort(AIsRecursive: Boolean);
Begin
  CustomSort(Nil, AIsRecursive);
End;

Procedure TZipTree.SetOnChange(AValue: TOnChangeEvent);
Var
  TI1: Integer;
Begin
  FOnChange := AValue;
  For TI1 := 0 To Count - 1 Do Items[TI1].OnChange := AValue; //Recursive Change
End;

End.

