unit VTHeaderPopup;

//----------------------------------------------------------------------------------------------------------------------
// The contents of this file are subject to the Mozilla Public License
// Version 1.1 (the "License"); you may not use this file except in
// compliance with the License. You may obtain a copy of the License at
// http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS"
// basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
// License for the specific language governing rights and limitations
// under the License.
//
// The Original Code is VTHeaderPopup.pas.
//
// The Initial Developer of the Original Code is Ralf Junker <delphi@zeitungsjunge.de>. All Rights Reserved.
//
// Modified 17 Feb 2002 by Jim Kueneman <jimdk@mindspring.com>.
//   Added the event to filter the items as they are added to the menu.
//
// Modified 23 Feb 2002 by Ralf Junker <delphi@zeitungsjunge.de>.
//   Added option to show menu items in the same order as the columns or in original order.
//   Added option to prevent the user to hide all columns.
//
// Modified 24 Feb 2002 by Ralf Junker <delphi@zeitungsjunge.de>.
//   Fixed a bug where the OnAddHeaderPopupItem would interfere with poAllowHideAll options.
//   All column indexes now consistently use TColumnIndex (instead of Integer).
//
//----------------------------------------------------------------------------------------------------------------------

{$I Compilers.inc}

interface

uses
  Menus, VirtualTrees;

type
  TVTHeaderPopupOption = (
    poOriginalOrder, // Show menu items in original column order as they were added to the tree.
    poAllowHideAll // Allows to hide all columns, including the last one.
  );
  TVTHeaderPopupOptions = set of TVTHeaderPopupOption;

  TAddPopupItemType = (
    apNormal,
    apDisabled,
    apHidden
  );

  TOnAddHeaderPopupItem = procedure(const Sender: TBaseVirtualTree; const Column: TColumnIndex; var Cmd: TAddPopupItemType) of object;

  TVTHeaderPopupMenu = class(TPopupMenu)
  private
    FOnAddHeaderPopupItem: TOnAddHeaderPopupItem;
    FOptions: TVTHeaderPopupOptions;
  protected
    procedure DoAddHeaderPopupItem(const Column: TColumnIndex; out Cmd: TAddPopupItemType);
    procedure OnMenuItemClick(Sender: TObject);
  public
    procedure Popup(x, y: Integer); override;
  published
    property OnAddHeaderPopupItem: TOnAddHeaderPopupItem read FOnAddHeaderPopupItem write FOnAddHeaderPopupItem;
    property Options: TVTHeaderPopupOptions read FOptions write FOptions;
  end;

//----------------------------------------------------------------------------------------------------------------------

implementation

uses
  Classes;

type
  TVirtualTreeCast = class(TBaseVirtualTree); // Necessary to make the header accessible.

//----------------- TVTHeaderPopupMenu ---------------------------------------------------------------------------------

procedure TVTHeaderPopupMenu.DoAddHeaderPopupItem(const Column: TColumnIndex; out Cmd: TAddPopupItemType);

begin
  Cmd := apNormal;
  if Assigned(FOnAddHeaderPopupItem) then
    FOnAddHeaderPopupItem(TVirtualTreeCast(PopupComponent), Column, Cmd);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TVTHeaderPopupMenu.OnMenuItemClick(Sender: TObject);

begin
  if PopupComponent = nil then
    Exit;

  with TMenuItem(Sender), TVirtualTreeCast(PopupComponent).Header.Columns.Items[Tag] do
    if Checked then
      Options := Options - [coVisible]
    else
      Options := Options + [coVisible];
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TVTHeaderPopupMenu.Popup(x, y: Integer);

var
  i: Integer;
  ColPos: TColumnPosition;
  ColIdx: TColumnIndex;

  NewMenuItem: TMenuItem;
  Cmd: TAddPopupItemType;

  VisibleCounter: Cardinal;
  VisibleItem: TMenuItem;
begin
  if PopupComponent = nil then Exit;

  // Delete existing menu items.
  i := Items.Count;
  while i > 0 do begin
      Dec(i);
      Items[i].Free;
    end;

  // Add column menu items.
  with TVirtualTreeCast(PopupComponent).Header do
    begin
      if hoShowImages in Options then Self.Images := Images;
      VisibleItem := nil;
      VisibleCounter := 0;
      for ColPos := 0 to Columns.Count - 1 do
        begin
          if poOriginalOrder in FOptions then
            ColIdx := ColPos
          else
            ColIdx := Columns.ColumnFromPosition(ColPos);
          with Columns[ColIdx] do
            begin
              if coVisible in Options then Inc(VisibleCounter);
              DoAddHeaderPopupItem(ColIdx, Cmd);
              if Cmd <> apHidden then
                begin
                  NewMenuItem := TMenuItem.Create(Self);
                  NewMenuItem.Tag := ColIdx;
                  NewMenuItem.Caption := Text;
                  NewMenuItem.Hint := Hint;
                  NewMenuItem.ImageIndex := ImageIndex;
                  NewMenuItem.Checked := coVisible in Options;
                  NewMenuItem.OnClick := OnMenuItemClick;
                  if Cmd = apDisabled then
                    NewMenuItem.Enabled := False
                  else
                    if coVisible in Options then VisibleItem := NewMenuItem;
                  Items.Add(NewMenuItem);
                end;
            end;
        end;
      // Conditionally disable menu item of last enabled column.
      if (VisibleCounter = 1) and (VisibleItem <> nil) and not (poAllowHideAll in FOptions) then
        VisibleItem.Enabled := False;
    end;
  inherited;
end;

//----------------------------------------------------------------------------------------------------------------------

end.

